import types
import inspect
from typing import Callable, Mapping, Sequence, List, Optional, Union, Literal, Tuple

type Funcs[T] = Callable[..., T] | Sequence[Callable[..., T]] | Mapping[str, Funcs[T]]


__version__: str

def run[T](
    funcs: Funcs[T],
    *,
    parsers: Mapping[type, Callable[[str], object]] = ...,
    short: Optional[Mapping[str, str]] = ...,
    cli_options: Literal["kwonly", "all", "has_default"] = ..., 
    show_defaults: bool = ..., 
    show_types: bool = ..., 
    no_negated_flags: bool = ..., 
    version: Union[str, None, bool] = ..., 
    argparse_kwargs: Mapping[str, object] = ...,
    intermixed: bool = ..., 
    argv: Optional[Sequence[str]] = ...,
) -> T: ...

def bind[T](
    funcs: Funcs[T],
    *,
    parsers: Mapping[type, Callable[[str], object]] = ...,
    short: Optional[Mapping[str, str]] = ...,
    cli_options: Literal["kwonly", "all", "has_default"] = ..., 
    show_defaults: bool = ..., 
    show_types: bool = ..., 
    no_negated_flags: bool = ..., 
    version: Union[str, None, bool] = ..., 
    argparse_kwargs: Mapping[str, object] = ...,
    intermixed: bool = ..., 
    argv: Optional[Sequence[str]] = ...,
) -> Callable[[], T]: ...

def bind_known[T](
    funcs: Funcs[T],
    *,
    parsers: Mapping[type, Callable[[str], object]] = ...,
    short: Optional[Mapping[str, str]] = ...,
    cli_options: Literal["kwonly", "all", "has_default"] = ..., 
    show_defaults: bool = ..., 
    show_types: bool = ..., 
    no_negated_flags: bool = ..., 
    version: Union[str, None, bool] = ..., 
    argparse_kwargs: Mapping[str, object] = ...,
    intermixed: bool = ..., 
    argv: Optional[Sequence[str]] = ...,
) -> Tuple[Callable[[], T], List[str]]: ...
class Parameter(inspect.Parameter):
    doc: Optional[str]
    def replace(
        self,
        *,
        doc: Optional[str] = ...,
        **kwargs: object,
    ) -> "Parameter": ...

class Signature(inspect.Signature):
    doc: Optional[str]
    raises: Tuple[type[BaseException], ...]

    @property
    def parameters(self) -> types.MappingProxyType[str, Parameter]: ...

    def replace(
        self,
        *,
        doc: Optional[str] = ...,
        raises: Tuple[type[BaseException], ...] = ...,
        **kwargs: object,
    ) -> "Signature": ...


def signature(func: Callable[..., object] | str) -> Signature: ...



from defopt import run, bind, bind_known, signature


def main(value: int, *, times: int = 1) -> int:
    """Print a value multiple times and return it.

    :param value: Value to display
    :param times: Number of repetitions
    :returns: The input ``value``
    """
    for _ in range(times):
        print(value)
    return value


def demo_bind() -> None:
    call = bind(main, argv=["1", "--times", "2"])
    ret1: int = call()
    call2, rest = bind_known(main, argv=["2", "--times", "3"])
    ret2: int = call2()


def demo_signature() -> None:
    sig = signature(main)
    doc: str | None = sig.doc
    param_doc: str | None = sig.parameters["value"].doc


def demo_run_dict() -> None:
    """Demonstrate running a mapping of callables.

    This ensures that ``run`` infers and returns the expected type when
    dispatching to a subcommand.
    """

    def double(value: int) -> int:
        return value * 2

    funcs = {"double": double}
    result: int = run(funcs, argv=["double", "3"])


if __name__ == "__main__":
    demo_bind()
    demo_signature()
    demo_run_dict()
    res: int = run(main)
